/* 
 * copyright 2010 Edscott Wilson Garcia (GPL-license)
 *
 * This is very simple example program to test 64 bit 
 * functions of the Disk Based Hash (DBH) and
 * verify correct handling of dbh files greater than
 * 2 Gb in size (up to 256^8/2).

 * A dbh file is created from a specified filesystem.
 * Paths are indexed with g_string hash key
 * Hash key collisions are noted in dbh file COLLISIONS
 * Hash key<->path associations are noted in dbh file KEY*
 * Hash key<->file are noted in dbh file INDEX
 *
 * usage: ./filesystem path option
 * Option can be:
 *    "index" (create KEY, COLLISIONS and INDEX dbh files)
 *    "dump"  (do a foreach on all records and print summary)
 *    "regen" (recreate INDEX dbh file with optimized fisical structure)
 *    "compare" (compare each file in INDEX with actual file on disk)
 *    "fulltest" (all of the above)

 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <dbh.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <glib.h>

#define KEY "filesystem.key.dbh"
#define INDEX "filesystem.index.dbh"
#define COLLISIONS "filesystem.collisions.dbh"

static long long sum=0;
static DBHashTable *dbh_key;
static
gchar *get_hash_key(unsigned char bucket, const char *pre_key){
    GString *gs = g_string_new(pre_key);
    gchar *key;
    key=g_strdup_printf("%c%10u", bucket, g_string_hash(gs));
    g_string_free(gs, TRUE);
    return key;
}

static int 
read_filesystem(DBHashTable *dbh, const char *path)
{
    DIR *directory; 
    int count = 0;
    struct dirent *d;

    directory = opendir(path);
    if(!directory) {
	fprintf(stderr,"Cannot open %s\n" ,path);
	return -1;
    }
#define     _BSD_SOURCE 1
while((d = readdir(directory)) != NULL)
    {

        char *fullpath=NULL;
	gchar  *key;
	gboolean is_dir=FALSE;
	unsigned char bucket='A';
        if(strcmp(d->d_name, ".")==0)  continue;
        if(strcmp(d->d_name, "..")==0)  continue;




        fullpath=(char *)malloc(strlen(path)+strlen("/")+strlen(d->d_name)+1);
	sprintf(fullpath,"%s/%s",path,d->d_name);
	if (strlen(fullpath) > 254){
	    printf("PATH too long(%d): %s\n",(int)strlen(fullpath),fullpath);
	    free(fullpath);
	    continue;
	}
	do {
	    key=get_hash_key(bucket,fullpath);
            dbh_set_key (dbh,(unsigned char *)key);
	    bucket++;
	}
	while(dbh_load(dbh));
	bucket--;
	if (bucket > 'A'){
	    printf("HASH colision: %s -> %s\n", key,fullpath);
	    DBHashTable *dbh_inverse=dbh_open(COLLISIONS);
	    char inverse_key[255];
	    memset(inverse_key,0,255);
	    strcpy(inverse_key,fullpath);
	    dbh_set_key (dbh_inverse,(unsigned char *)inverse_key);
	    dbh_set_size(dbh_inverse,DBH_KEYLENGTH(dbh));
	    dbh_set_data(dbh_inverse,(void *)DBH_KEY(dbh),DBH_KEYLENGTH(dbh));
	    dbh_update(dbh_inverse);
	    dbh_close(dbh_inverse);
	} 

	dbh_set_key (dbh_key,(unsigned char *)key);
	dbh_set_size(dbh_key,strlen(fullpath)+1);
	dbh_set_data(dbh_key,(void *)fullpath,strlen(fullpath)+1);
	dbh_update(dbh_key);
	 struct stat st;
	 if (lstat(fullpath,&st)<0){
	    printf("cannot stat %s: %s\n",fullpath,strerror(errno));
	    continue;
	 }


	if (S_ISDIR(st.st_mode)) is_dir=TRUE;

        if (!is_dir) {

	 if (st.st_size == 0) continue; 
	 if (st.st_size > 100000000LL) { 
	     printf("file too big (%lld): %s\n",(long long)st.st_size,fullpath);
	     continue;
	 }
	 if (!S_ISREG(st.st_mode)) {
	     continue;
	 }
	 // This is useful if our data size grows over 1024 B:
	 if (DBH_MAXIMUM_RECORD_SIZE(dbh) < st.st_size) {
	     dbh_set_size(dbh,st.st_size);
	     printf("dbh_set_size set to %lld\n",(long long)st.st_size);
	 }
	 int fd=open(fullpath,O_RDONLY);
	 if (fd < 0) {
	    printf("cannot open %s for read\n",fullpath);
	    continue;
	 }
	 // This works instead of dbh_set_data():
	 if (read(fd,DBH_DATA(dbh),st.st_size) < 0){
	    printf("problem reading %lld bytes from %s\n",
		    (long long)st.st_size,fullpath);
	    close(fd);
	    continue;
	 }
	 close(fd);
	 dbh_set_recordsize(dbh,st.st_size);
	 dbh_update(dbh);
	 sum += st.st_size;
	 count++;
         //fprintf(stderr,"%s\n",fullpath);
	 //
	}
        if (is_dir) {
		int retval;
		retval = read_filesystem(dbh,fullpath);
		if (retval > 0) count += retval;
	}

	free(fullpath);
    }
    closedir(directory);
//	printf ("%s -> %d files\n",path,count);
    return (count);
}

static int count2=0;
void  operate (DBHashTable *dbh){
    count2++;
    //sum += strlen((char *)DBH_DATA(dbh));
    sum += DBH_RECORD_SIZE(dbh);
}
void  compare (DBHashTable *dbh){
    dbh_set_key (dbh_key,(unsigned char *)DBH_KEY(dbh));
    dbh_load(dbh_key);
    char *path=DBH_DATA(dbh_key);
    int fd=open(path,O_RDONLY);
	 if (fd < 0) {
	    printf("cannot open %s for read\n",path);
	    return;
	 }
	 // This works instead of dbh_set_data():
    struct stat st;
    lstat(path,&st);
    void *p=malloc(st.st_size);
	 if (read(fd,p,st.st_size) < 0){
	    printf("problem reading %lld bytes from %s\n",
		    (long long)st.st_size,path);
	    close(fd);
	    free(p);
	    return;
	 }
	 close(fd);
    if (memcmp(p,DBH_DATA(dbh),st.st_size) != 0) {
	printf("%s does not compare!\n",path);
    } else {
	static int count=0;
	if (count++ % 1000 == 0) {
	    printf ("."); fflush(stdout);
	}
    }
    free(p);
}
int main(int argc, char **argv){
  if (argc < 3) {
   fprintf(stderr,"insufficient arguments (%d), use: %s path index|dump|compare|regen|fulltest\n",argc,argv[0]);
   exit(1);
  }
    
  if (strcmp(argv[2],"index")==0 || strcmp(argv[2],"fulltest")==0) {
    fprintf(stdout,"Creating index now\n");
    DBHashTable *dbh=dbh_create(INDEX,11);
    dbh_key=dbh_create(KEY,11);
    DBHashTable *dbh_inverse=dbh_create(COLLISIONS,255);
    dbh_close(dbh_inverse);
    int count=read_filesystem(dbh,argv[1]);
    dbh_close(dbh);
    dbh_close(dbh_key);
    fprintf(stdout,"Filesystem count=%d sum=%lld\n",count,sum);
  }
  if (strcmp(argv[2],"dump")==0 || strcmp(argv[2],"fulltest")==0) {
    count2=0;   sum=0;
    fprintf(stdout,"Dumping now\n");
    DBHashTable *dbh=dbh_open(INDEX);
    dbh_foreach_sweep (dbh,operate);
    fprintf(stdout,"Filesystem dbh count=%d sum=%lld\n",count2,sum);
    dbh_close(dbh);
  }
  if (strcmp(argv[2],"regen")==0 || strcmp(argv[2],"fulltest")==0) {
    fprintf(stdout,"Regen_sweep now\n");
    DBHashTable *dbh=dbh_open(INDEX);
    dbh=dbh_regen_sweep(dbh);
    count2=0;   sum=0;
    dbh_foreach_sweep (dbh,operate);
    fprintf(stdout,"dbh_foreach_sweep count=%d sum=%lld\n",count2,sum);
    fprintf(stdout,"Regen_fanout now\n");
    dbh=dbh_regen_fanout(dbh);
    count2=0;   sum=0;
    dbh_foreach_fanout (dbh,operate);
    dbh_close(dbh);
    fprintf(stdout,"dbh_foreach_fanout count=%d sum=%lld\n",count2,sum);
  }
  
  if (strcmp(argv[2],"compare")==0 || strcmp(argv[2],"fulltest")==0) {
    fprintf(stdout,"Comparing now\n");
    DBHashTable *dbh=dbh_open(INDEX);
    dbh_key=dbh_open(KEY);
    dbh_foreach_sweep (dbh,compare);
    dbh_close(dbh);
    dbh_close(dbh_key);
  }
  return 1;
}
