/***************************************************************************

    Disk Based Hash library DBH


    Header file.


copyright Edscott Wilson Garcia 2002-2010
published under LGPL Public license
You should have received a license copy along with this file.

notice: MODIFICATIONS TO THIS FILE "dbh.h" WILL BE LOST. 
             MODIFY FILE "dbh.h.in" INSTEAD.
***************************************************************************/
#ifndef DBH_H
#define DBH_H

#undef DBH_VERSION
/**
 * DBH_VERSION
 *
 * Disk Based Hashtables library version
 *
 **/
#define DBH_VERSION 		"4.6.1"

#undef DBH_FILE_VERSION
/**
 * DBH_FILE_VERSION
 *
 * Disk Based Hashtables library file version compatibility
 *
 **/
#define DBH_FILE_VERSION 	"DBH_2.0/64bit"

#undef FILE_POINTER
/**
 * FILE_POINTER: 
 * 
 * Architecture independent 64 bit integer type  
 *
 **/
#define FILE_POINTER 	\
    long long

/**
 * SECTION:dbh
 * @short_description: 	Library to create and manage hash tables residing 
 * 			on disk. Associations are made between keys and 
 * 			values so that for a given a key the value can be 
 * 			found and loaded into memory quickly. Being disk based 
 * 			allows for large and persistent hashes. 64 bit support
 * 			allows for hashtables with sizes over 4 Gigabytes on 32
 * 			bit systems. Cuantified key generation allows for 
 * 			minimum access time on balanced multidimensional trees.
 * @title: 		Disk Based Hashtables (DBH) 64 bit
 * @section_id:		DBH
 * @see_also: #GHashTables
 * @stability: Stable
 * @include: dbh.h
 * @Image: calc.png
 *
 * <para>
 * A DBHashTable  provides associations between keys and values which is 
 * optimized so that given a key, the associated value can be found very 
 * quickly.
 * </para>
 * <para>
 * Note that only one hash record is loaded from disk to memory at any 
 * given moment for a DBHashTable. Both keys and values should be copied 
 * into the DBHashTable record, so they need not exist for the lifetime 
 * of the DBHashTable. This means that the use of static strings and 
 * temporary strings (i.e. those created in buffers and those returned by
 * GTK+ widgets) should be copied with dbh_set_key() and dbh_set_data()
 * into the DBHashTable record before being inserted.
 * </para>
 * <para>
 * You must be careful to ensure that copied key length matches the defined 
 * key length of the DBHashTable, and also that the copied data does not 
 * exceed the maximum length of the DBHashTable record (1024 bytes by 
 * default, and expandable by dbh_set_size() ). If the DBHashTable record length 
 * is to be variable, be sure to set the appropriate length before each 
 * dbh_update(), with dbh_set_recordsize(), otherwise the record length 
 * need only be set before the first dbh_update().
 * </para>
 * <para>
 * To create a DBHashTable, use dbh_create().
 * </para>
 * <para>
 * To insert a key and value into a DBHashTable, use dbh_update(). 
 * The DBHashTable will not be modified until this command is given. 
 * All changes to the current DBHashTable record only reside in memory. 
 * dbh_update() is necessary to commit the changes to the DBHashTable.
 * </para>
 * <para>
 * To lookup a value corresponding to a given key, use dbh_load().
 * </para>
 * <para>
 * To erase and unerase a key and value, use dbh_erase() and dbh_unerase().
 * </para>
 * <para>
 * To call a function for each key and value pair (using a sweep route) 
 * use dbh_foreach_sweep() and dbh_sweep().
 * </para>
 * <para>
 * To call a function for each key and value pair (using a fanout route) 
 * use dbh_foreach_fanout() and dbh_foreach_fanout().
 * </para>
 * <para>
 * To destroy a DBHashTable use dbh_destroy().
 * </para>
 * <para>
 * This is dbh version 2, incompatible with dbh version 1 files. 
 * The main difference between the two version is the handling of 
 * file pointers. In version 1, file pointers were 32 bits in length, 
 * while in version 2, file pointers are 64 bits in length. 
 * This allows for DBHashTables with sizes greater than 2 GBytes.
 * </para>
 * <para>
 * <quote>Cuantified numbers</quote> are an alternate way to view 
 * the set of <quote>natural numbers</quote> {1, 2, 3, ...} where
 * order is defined in two levels. 
 * In <quote>natural numbers</quote> there is only one level of order
 * (defined by the > boolean operator). In 
 * <quote>cuantified numbers</quote> 
 * the first level of order is defined by the <quote>cuanta</quote> 
 * or quantity. The <quote>cuanta</quote> is obtained by adding all 
 * the digits of the <quote>cuantified number</quote>.
 * Thus, for example, 10022, 5, 32, and 11111 are all equal at the
 * first level of order since they all add up to 5. The second level
 * or order may be obtained in different manners. In functions dbh_genkey()
 * and dbh_genkey2() the corresponding order of the
 * <quote>natural numbers</quote> from which they are associated is 
 * <emphasis>not conserved</emphasis>.
 * </para>
 * <para>
 * In dbh_orderkey() the corresponding order of the
 * <quote>natural numbers</quote>  from which they are associated
 * <emphasis>is conserved</emphasis>, but at a price. 
 * The base, or maximum value each digit may reach, must be defined. 
 * This effectively puts a limit on the number of keys which may be 
 * generated for a given number of digits.
 * </para>
 * <para>
 * When a #DBHashTable is constructed with <quote>cuantified</quote> keys, the 
 * maximum amount of disk access instructions generated to access
 * any given record is equal to the <quote>cuanta</quote> of the cuantified number
 * represented by the key. This allows a #DBHashTable to be constructed with 
 * minimum access time across all records.
 * </para>

 **/

/* NOTICE: dbh_header_t structure MUST be data aligned!  */
/**
 * dbh_header_t:
 * @n_limit:  Maximum toplevel branches 
 * @user_chars: Five unsigned chars available to user
 * @bof: File pointer to root of tree
 * @erased_space: Amount of bytes marked as erased
 * @data_space: Amount of bytes ocuppied by data
 * @total_space: Amount of bytes ocuppied by data and format
 * @records: Number of records
 * @record_length: Maximum record length
 * @user_filepointer: Six 64-bit filepointers available to user
 * @version: DBHashTable version compatibility information
 * @copyright: DBH sourcecode distribution copyright and download information
 *
 * #dbh_header_t is the structural information written at the first 256 bytes of
 * a #DBHashTable file.
 * 
 **/
typedef struct _dbh_header_t dbh_header_t;
struct _dbh_header_t
{
/*< private >*/
	/****    16/256 bytes  *****/
/*< public >*/
  unsigned char n_limit;	
/*< private >*/
  unsigned char position;		// position of key in data 
  unsigned char length;			// length of key in data 
  unsigned char totalkeylength;		// total length of key in data 
  unsigned char user_tmpdir;		// volatile (boolean)
  unsigned char DBH32;			// not used 
  unsigned char reservedC;		// used by system, value goes to 0 on updates or inserts) 
  unsigned char reservedD;		// same as above, but not used anymore 
  unsigned char sweep_erased;		// 0
  unsigned char writeOK;		// reserved for inner use 
  unsigned char dbh_exit;		// reserved for inner use 
/*< public >*/
  unsigned char user_chars[5];	
/*< private >*/
	/****     12*8 = 96/256 bytes  *****/
/*< public >*/
  FILE_POINTER bof;		
  FILE_POINTER erased_space;	
  FILE_POINTER data_space;	
  FILE_POINTER total_space;	
				   
  FILE_POINTER records;		
  FILE_POINTER record_length;	
  FILE_POINTER user_filepointer[6];	
/*< private >*/
	/****     16+128 = 144/256 bytes  *****/
/*< public >*/
  char version[16];		
  char copyright[128];		
/*< private >*/
};

/**
 * DBHashTable:
 * @branches: Maximum toplevel branches
 * @bytes_userdata: size of data record
 * @key: access key
 * @data: record data pointer
 * @fd: file descriptor
 * @head_info: nonvolatile header information
 * @path: file path
 * 
 *
 * #DBHashTable is a data structure containing the record information for an open
 * #DBHashTable file.
 *
 */
typedef struct _DBHashTable DBHashTable;
struct _DBHashTable
{
/*< public >*/
  unsigned char branches;		    
/*< private >*/
  unsigned char newbranches;		// inner use (find)   
  unsigned char flag;			// reserved (erase)     
  FILE_POINTER reservedB;		// reserved             
/*< public >*/
  FILE_POINTER bytes_userdata;		   
/*< private >*/
  FILE_POINTER newbytes_userdata;	// inner use (find)  
  FILE_POINTER *branch;			// inner use (format)
  FILE_POINTER *newbranch;		// inner use (find) 
/*< public >*/
  unsigned char *key;			 
/*< private >*/
  unsigned char *newkey;
  unsigned char *reservedC;		// inner use 
  unsigned char *reservedD;		// inner use 
/*< public >*/
  void *data;				 
/*< private >*/
  void *newdata;			// inner use (find) 
  void (*operate) (struct _DBHashTable *);// para funcion barre
  void (*reservedE) (void *);		// reserved 
  void *((*reservedF) (void *));	// reserved  
  void *((*reservedG) (void *));	// reserved  
/*< public >*/
  int fd;				 
  dbh_header_t *head_info;		 
  char *path;				
/*< private >*/
  char *tmpdir;
  struct _DBHashTable *dbh_desdbh;

} ;
/////////   macros     ///////////////////////////////////////
/**
 * DBH_KEYLENGTH:
 * @dbh: 	A #DBHashTable pointer (#DBHashTable *)
 *
 * 		This macro returns the keylenth in bytes associated
 * 		to the #DBHashTable. The value is fixed when the 
 *		#DBHashTable is created with #dbh_create.
 *
 **/
#define DBH_KEYLENGTH(dbh) \
				(dbh->head_info->n_limit)
/**
 * DBH_RECORD_SIZE:
 * @dbh: 	A #DBHashTable pointer (#DBHashTable *)
 *
 * 	 	This macro returns the size of the current record loaded in 
 * 	 	memory. If no record has been loaded, then the return value 
 * 		is not defined.
 *
 **/
#define DBH_RECORD_SIZE(dbh) 	\
				(dbh->bytes_userdata)
/**
 * DBH_KEY:
 * @dbh: 	A #DBHashTable pointer (#DBHashTable *)
 * 
 * 	 	This macro returns a pointer to the current 
 * 	 	#DBHashTable key area. 
 *
 **/
#define DBH_KEY(dbh) 		\
				(dbh->key)

/**
 * DBH_DATA:
 * @dbh: 	A #DBHashTable pointer (#DBHashTable *)
 *
 * 	 	This macro returns a pointer to the current 
 * 	 	#DBHashTable data area.
 *
 **/
#define DBH_DATA(dbh) 		\
				(dbh->data)
/**
 * DBH_ERASED_SPACE:
 * @dbh: 	A #DBHashTable pointer (#DBHashTable *)
 *
 * 	 	This macro returns  the amount of 
 * 	 	bytes taken up by erased data in the #DBHashTable. 
 *
 **/
#define DBH_ERASED_SPACE(dbh) 	\
				(dbh->head_info->erased_space)
/**
 * DBH_DATA_SPACE:
 * @dbh: 	A #DBHashTable pointer (#DBHashTable *)
 *
 * 	 	This macro returns  the amount of
 * 	 	bytes taken up by valid data in the #DBHashTable.  
 *
 **/
#define DBH_DATA_SPACE(dbh) 	\
				(dbh->head_info->data_space)
/**
 * DBH_TOTAL_SPACE:
 * @dbh: 	A #DBHashTable pointer (#DBHashTable *)
 * 
 * 	 	This macro returns  the total amount 
 * 	 	of bytes taken up by the #DBHashTable.  
 *
 **/
#define DBH_TOTAL_SPACE(dbh) 	\
				(dbh->head_info->total_space)
/**
 * DBH_FORMAT_SPACE:
 * @dbh: 	A #DBHashTable pointer (#DBHashTable *)
 *
 * 	 	This macro returns the total amount 
 * 	 	of bytes taken up by the format of the #DBHashTable. 
 *
 **/
#define DBH_FORMAT_SPACE(dbh) 	\
				(dbh->head_info->total_space - dbh->head_info->data_space - dbh->head_info->erased_space)
/**
 * DBH_RECORDS:
 * @dbh: 	A #DBHashTable pointer (#DBHashTable *)
 *
 * 	 	This macro returns the number of records in the #DBHashTable.  
 *
 **/
#define DBH_RECORDS(dbh) 	\
				(dbh->head_info->records)
/**
 * DBH_MAXIMUM_RECORD_SIZE:
 * @dbh: 	A #DBHashTable pointer (#DBHashTable *)
 *
 * 	 	This macro returns the maximum allocated space for 
 * 	 	data in the current #DBHashTable record.  
 *
 **/
#define DBH_MAXIMUM_RECORD_SIZE(dbh) \
				    (dbh->head_info->record_length)
/**
 * DBH_PATH:
 * @dbh: 	A #DBHashTable pointer (#DBHashTable *)
 *
 * 	 	This macro returns a pointer to a string containing 
 * 	 	the path to the current #DBHashTable.
 *
 **/
#define DBH_PATH(dbh)		\
				(dbh->head_info->archivo)


//////////  functions  ///////////////////////////////////////
/**
 * DBHashFunc:
 * @dbh: A #DBHashTable pointer (#DBHashTable *)
 *
 * Pointer to function to apply during dbh_sweep(), dbh_fanout(),
 * dbh_foreach_sweep() and dbh_foreach_fanout().
 * <para>
 * This function will be applied to all data records involved 
 * in the sweep or fanout process
 * </para>
 *
 **/
typedef void (*DBHashFunc) (DBHashTable *dbh);
/* level 1 functions ***********************************/

int dbh_close (DBHashTable * dbh);
int dbh_destroy (DBHashTable * dbh);

DBHashTable *dbh_create (const char *path, unsigned char key_length);
DBHashTable *dbh_open (const char *path);
DBHashTable *dbh_open_ro (const char *path);

int dbh_erase (DBHashTable * dbh);
int dbh_unerase (DBHashTable * dbh);
int dbh_prune (DBHashTable * dbh,
	       unsigned char *key, unsigned char subtree_length);
int dbh_unprune (DBHashTable * dbh,
		 unsigned char *key, unsigned char subtree_length);

FILE_POINTER dbh_find (DBHashTable * dbh, int n);

void dbh_genkey (unsigned char *key, unsigned char length, unsigned int n);
void dbh_genkey2 (unsigned char *key, unsigned char length, unsigned int n);
void dbh_orderkey (unsigned char *key,
	      unsigned char length, unsigned int n, unsigned char base);

FILE_POINTER dbh_load (DBHashTable * dbh);
unsigned char dbh_load_address (DBHashTable * dbh, FILE_POINTER currentseek);
FILE_POINTER dbh_load_parent (DBHashTable * dbh);
FILE_POINTER dbh_load_child (DBHashTable * dbh, unsigned char key_index);

DBHashTable *dbh_regen_sweep (DBHashTable * dbh);
DBHashTable *dbh_regen_fanout (DBHashTable * dbh);
int dbh_settempdir (DBHashTable * dbh, char *temp_dir);

void dbh_set_data (DBHashTable * dbh, void *data, FILE_POINTER size);
void dbh_set_key (DBHashTable * dbh, unsigned char *key);

int dbh_set_size (DBHashTable * dbh, FILE_POINTER size);
void dbh_set_recordsize (DBHashTable * dbh, int record_size);

int
dbh_sweep (DBHashTable * dbh,
	   DBHashFunc operate,
	   unsigned char *key1,
	   unsigned char *key2, unsigned char ignore_portion);
int
dbh_fanout (DBHashTable * dbh,
	    DBHashFunc operate,
	    unsigned char *key1,
	    unsigned char *key2, unsigned char ignore_portion);
int dbh_foreach_sweep (DBHashTable * dbh, DBHashFunc operate);
int dbh_foreach_fanout (DBHashTable * dbh, DBHashFunc operate);
void dbh_exit_sweep (DBHashTable * dbh);
void dbh_exit_fanout (DBHashTable * dbh);

FILE_POINTER dbh_update (DBHashTable * dbh);
int dbh_writeheader (DBHashTable * dbh);

#endif
