/***************************************************************************

    Disk Based Hash library DBH

    exported functions.

copyright Edscott Wilson Garcia 2002-2010
published under LGPL Public license
You should have received a license copy along with this file.

*************************************************************************/
/**
 * dbh_create:
 * @path: Path on disk where DBHashTable will reside.
 * @key_length: The length of the key to access the DBHashTable. 
 * @Returns: 	A pointer to the newly created and opened #DBHashTable, 
 * 		or NULL if it fails.
 *
 * 		Create a new hash file (overwriting old version). 
 * 		Creates and opens for writing a new #DBHashTable. 
 * 		This function will overwrite any file with the specified
 * 		path, including any previous DBH file. The @key_length is
 * 		fixed. If you want variable length, use a g_hash_table 
 * 		to associate cuantified keys generated by genkey(), and 
 * 		create an extra DBHashTable to save the g_hash. 
 * 		Cuantified  keys assure that large DBHashes are spread out
 * 		optimally.
 **/
DBHashTable *
dbh_create (const char *path, unsigned char key_length)
{
  DBHashTable *dbh;
  dbh = (DBHashTable *) malloc (sizeof (DBHashTable));
  if (dbh == NULL)
    return NULL;
  memset (dbh, 0, sizeof (DBHashTable));
  unlink(path);
  dbh->fd = open(path, O_CREAT|O_RDWR|O_TRUNC, 0640);
  if (dbh->fd < 0)
    {
      free (dbh);
      fprintf(stderr,"dbh_create(%s): %s\n",path, strerror(errno));
      return NULL;
    }
  dbh->head_info = (dbh_header_t *) malloc (sizeof (dbh_header_t));
  memset ((void *) (dbh->head_info), 0, sizeof (dbh_header_t));

  dbh->head_info->bof = sizeof (dbh_header_t);
  dbh->head_info->n_limit = key_length;
  dbh->head_info->user_tmpdir = 0;
  dbh->head_info->sweep_erased = 0;
  dbh->head_info->total_space = dbh->head_info->data_space =
    dbh->head_info->erased_space = 0;
  strncpy (dbh->head_info->version, DBH_FILE_VERSION, 15);
  dbh->head_info->records = 0;
  dbh->operate = dbh_operate;
  dbh->branch =
    (FILE_POINTER *) malloc (dbh->head_info->n_limit *
			     sizeof (FILE_POINTER));
  dbh->newbranch =
    (FILE_POINTER *) malloc (dbh->head_info->n_limit *
			     sizeof (FILE_POINTER));
  dbh->key = (unsigned char *) malloc (key_length);
  dbh->newkey = (unsigned char *) malloc (key_length);
  dbh->path = (char *) malloc (strlen (path) + 1);
  strcpy (dbh->path, path);

  dbh->head_info->record_length = 0;
  dbh->head_info->writeOK = 1;
  dbh_size (dbh, DEFAULT_DBH_DATASIZE);
  dbh->bytes_userdata = 0;
  dbh_writeheader (dbh);
  DBG ("DBH: created %s \n", path);
  return dbh;
}  

/**
 * dbh_open:
 * @path: Path on disk where DBHashTable resides.
 * @Returns: A pointer to the newly opened #DBHashTable, 
 * 		or NULL if it fails.
 *
 * Open an existing hash in read-write mode. A file write lock will be placed
 * on the file until dbh_close() is called.
 *
 **/
DBHashTable *
dbh_open (const char *path)
{
  return (dbh_open_S (path, WRITE));
}

/**
 * dbh_open_ro:
 * @path: Path on disk where DBHashTable resides.
 * @Returns: A pointer to the newly opened read-only DBHashTable, 
 * 		or NULL if it fails.
 *
 * Open an existing hash in read-only mode. A file read lock will be placed
 * on the file until dbh_close() is called.
 *
 **/
DBHashTable *
dbh_open_ro (const char *path)
{
  return (dbh_open_S (path, READ));
}

/**
 * dbh_close:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @Returns: 0 on error, 1 otherwise.
 *
 * Close hash file (thus flushing io buffer).
 *
 **/
int
dbh_close (DBHashTable * dbh)
{
  if (dbh == NULL){
      fprintf(stderr,"dbh_close(dbh): dbh == NULL\n");
    return 0;
  }
  if (dbh->head_info->writeOK)
    dbh_writeheader (dbh);

  if (close (dbh->fd) < 0) {
      fprintf(stderr,"dbh_close(): %s\n",strerror(errno));
  }

  free (dbh->data);
  free (dbh->newdata);
  free (dbh->branch);
  free (dbh->newbranch);
  free (dbh->key);
  free (dbh->newkey);
  free (dbh->head_info);
  free (dbh->path);
  free (dbh);
  return 1;
} 

/**
 * dbh_destroy:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @Returns: 0 if error, 1 otherwise
 *
 * Close an open DBHashTable and erase file from disk.
 * Convenience function that does a close and rm.
 *
 **/
int
dbh_destroy (DBHashTable * dbh)
{
  char *file;
  if (!dbh)
    {
      fprintf (stderr, "dbh_destroy(): %s\n", strerror (EBADF));

      return ERROR_VALUE;
    }

  file = (char *) malloc (strlen (dbh->path) + 1);
  strcpy (file, dbh->path);
  dbh_close (dbh);
  int retval = remove (file);

  if (retval < 0)
    retval = ERROR_VALUE;
  else
    retval = 1;
  free (file);
  return 1;
}

/**
 * dbh_set_size:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @size: size in bytes.
 * @Returns: 0 on error, 1 otherwise.
 *
 * Defines the maximum amount of memory to be allocated to the 
 * #DBHashTable records. This is nonvolatile information which
 * need to be set only once. The default is 1Kbyte. 
 *
 **/
int
dbh_set_size (DBHashTable * dbh, FILE_POINTER record_length)
{
  return (dbh_size (dbh, record_length));
}

/**
 * dbh_set_recordsize:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @record_size: The amount of bytes in the current #DBHashTable record.
 *
 * This sets the recordsize of the the data in the current #DBHashTable 
 * record. It is called implicitly by calling dbh_set_data(). It is very 
 * important to call this function. Unpredictable results will follow if
 * record_size is not set. #DBHashTable records are variable in length, so 
 * use this function at least once if you are planning to use fixed length
 * records. This function is not needed if dbh_set_data() is used to set
 * the record data.
 *
 **/
void
dbh_set_recordsize (DBHashTable * dbh, int size)
{
  if (!dbh)
    {
      fprintf (stderr, "dbh_set_recordsize(): %s\n", strerror (EBADF));

      return;
    }

  dbh->bytes_userdata = size;
}

    
/**
 * dbh_set_data:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @data: Pointer to the data to copy to the current #DBHashTable record
 * @size: The amount of bytes to copy to the current #DBHashTable record
 *
 * This function copies the user data into the current #DBHashTable record 
 * and along with function dbh_set_key(), makes the current #DBHashTable 
 * record ready for the dbh_update() function to commit to the actual 
 * #DBHashTable on disk. 
 *
 **/
void
dbh_set_data (DBHashTable * dbh, void *data, FILE_POINTER n)
{
  if (!dbh || !data)
    {
      fprintf (stderr,"DBH: invalid parameter in dbh_set_data()\n");
      return;
    }
  if (n > dbh->head_info->record_length)
    {
      fprintf (stderr,"DBH: redefining maximum record size to %lld\n",(long long) n);
      dbh_set_size (dbh, n);
      return;
    }
  memcpy ((void *) dbh->data, (void *) data, n);
  dbh->bytes_userdata = n;
}

/**
 * dbh_set_key:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @key: The key to set as the current DBHashTable record key.
 *
 * This function sets the key of the current DBHashTable record.
 *
 **/
void
dbh_set_key (DBHashTable * dbh, unsigned char *key)
{
  if (!dbh || !key)
    {
      fprintf (stderr,"DBH: invalid parameter in dbh_set_key()\n");
      return;
    }
  memcpy ((void *) dbh->key, (void *) key, dbh->head_info->n_limit);

}

/**
 * dbh_update:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @Returns: 0 on error, byte offset of loaded record otherwise.
 *
 * Update the current record in memory to the disk based hash. Update function 
 * will update erased records as well as unerased records, but if
 * an erased record is updated, it is automatically unerased.
 *
 **/
FILE_POINTER
dbh_update (DBHashTable * dbh)
{
  TRACE ("dbh_update\n");
  FILE_POINTER fp[3];
  unsigned char j, caso;
  int i;

  if (dbh == NULL)
    return ERROR_VALUE;

  /* before updating, clean erased flag, so updated record will be 
   * automatically unerased this allows reusing the erased space, 
   * although it introduces error in the tabulated values of 
   * erased_space and total_space, but who cares, it's an 
   * approximation anyways. */
  SET_UNERASED;

  dbh->head_info->reservedC = 0;


  if (dbh_locate (dbh, fp) == NULL)
    return ERROR_VALUE;

  if (CURRENTSEEK)
    {
      if (dbh->newbytes_userdata < dbh->bytes_userdata)
	{
	  if (LASTSEEK)
	    caso = PRESENTE_MENOR;
	  else
	    caso = PRESENTE_MENOR_BOF;
	}
      else
	caso = PRESENTE_MAYORIGUAL;
    }
  else
    {
      if (LASTSEEK)
	caso = NO_PRESENTE;
      else
	caso = ARCHIVO_VACIO;
    }
  dbh->flag = 0;
  switch (caso)
    {
    case NO_PRESENTE:		/* not found */
      for (i = 0; i < dbh->head_info->n_limit; i++)
	dbh->branch[i] = 0;
      CURRENTSEEK = place_eof (dbh);
      if (CURRENTSEEK < 0)
	return ERROR_VALUE;
      dbh->newbranches -= (unsigned char) CURR_BRANCH;
      dbh->head_info->data_space += dbh->bytes_userdata;
      dbh->head_info->total_space +=
	(dbh->bytes_userdata + sizeof (FILE_POINTER) * dbh->newbranches +
	 1 + sizeof (FILE_POINTER));
      if (!dbh_write (OLD, dbh, WRITEBRANCHES))
	return ERROR_VALUE;
      if (!dbh_readBranches (dbh, LASTSEEK))
	return ERROR_VALUE;
      dbh->newbranch[CURR_BRANCH] = CURRENTSEEK;
      dbh_updateBranch (dbh, LASTSEEK);
      dbh->head_info->records++;
      break;	/*********************************/
    case PRESENTE_MAYORIGUAL:	/* use already assigned disk space */
      dbh->head_info->erased_space +=
	(dbh->newbytes_userdata - dbh->bytes_userdata);
      dbh->head_info->data_space -=
	(dbh->newbytes_userdata - dbh->bytes_userdata);
      dbh_lock_write (dbh);
      if (!place_fp_at (dbh, CURRENTSEEK))
	return ERROR_VALUE;
      if (!dbh_write (OLD, dbh, DONTWRITEBRANCHES))
	{
	  dbh_unlock (dbh);
	  return ERROR_VALUE;
	}
      dbh_unlock (dbh);
      break;	/*********************************/
    case PRESENTE_MENOR_BOF:	/* does not fit in assigned disk space and 
				   is the first record (bof) to boot */
      dbh->head_info->erased_space += (dbh->newbytes_userdata);
      dbh->head_info->data_space +=
	(dbh->bytes_userdata - dbh->newbytes_userdata);
      dbh->head_info->total_space +=
	(dbh->bytes_userdata + sizeof (FILE_POINTER) * dbh->newbranches +
	 1 + sizeof (FILE_POINTER));
      dbh_lock_write (dbh);
      FILE_POINTER eof = place_eof (dbh);
      if (eof < 0)
	{
	  dbh_unlock (dbh);
	  return ERROR_VALUE;
	}
      dbh->head_info->bof = eof;
      if (!dbh_write (NEW, dbh, WRITEBRANCHES))
	{
	  dbh_unlock (dbh);
	  return ERROR_VALUE;
	}
      dbh_writeheader (dbh);
      dbh_unlock (dbh);
      break;	/*********************************/
    case PRESENTE_MENOR:	/* does not fit in assigned disk space */
      {
	unsigned char ramas;
	dbh->head_info->erased_space += (dbh->newbytes_userdata);
	dbh->head_info->data_space +=
	  (dbh->bytes_userdata - dbh->newbytes_userdata);
	dbh->head_info->total_space +=
	  (dbh->bytes_userdata + sizeof (FILE_POINTER) * dbh->newbranches +
	   1 + sizeof (FILE_POINTER));
	dbh_lock_write (dbh);
	CURRENTSEEK = place_eof (dbh);
	if (CURRENTSEEK < 0)
	  {
	    dbh_unlock (dbh);
	    return ERROR_VALUE;
	  }
	j = dbh->newbranches;
	if (!dbh_write (NEW, dbh, WRITEBRANCHES))
	  {
	    dbh_unlock (dbh);
	    return ERROR_VALUE;
	  }
	ramas = dbh_readBranches (dbh, LASTSEEK);
	if (!ramas)
	  {
	    dbh_unlock (dbh);
	    return ERROR_VALUE;
	  }
	dbh->newbranch[ramas - j + CURR_BRANCH] = CURRENTSEEK;
	dbh_updateBranch (dbh, LASTSEEK);
      }
      dbh_unlock (dbh);
      break;	 /*********************************/
    case ARCHIVO_VACIO:  /* empty dbh file */
      for (i = 0; i < dbh->head_info->n_limit; i++)
	dbh->branch[i] = 0;
      CURRENTSEEK = dbh->head_info->bof;
      dbh_lock_write (dbh);
      if (!place_fp_at (dbh, CURRENTSEEK))
	{
	  dbh_unlock (dbh);
	  return ERROR_VALUE;
	}
      dbh->newbranches = dbh->head_info->n_limit;
      dbh->head_info->data_space += dbh->bytes_userdata;
      dbh->head_info->total_space +=
	(dbh->bytes_userdata + sizeof (FILE_POINTER) * dbh->newbranches +
	 1 + sizeof (FILE_POINTER));
      if (!dbh_write (OLD, dbh, WRITEBRANCHES))
	{
	  dbh_unlock (dbh);
	  return ERROR_VALUE;
	}
      dbh->head_info->records++;
      dbh_unlock (dbh);
      break;	/*********************************/
    }
  return CURRENTSEEK;
}   

/**
 * dbh_load:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @Returns: 0 on error, byte offset of loaded record otherwise.
 *
 * Load a record using the currently set key. This function will also
 * load erased values, except that it will return 0. 
 * 
 **/
FILE_POINTER
dbh_load (DBHashTable * dbh)
{
  int i;
  int j;
  FILE_POINTER fp[3];
  unsigned char *tmp1, *tmp2;
  if (dbh == NULL)
    return ERROR_VALUE;

  /* before loading, clean erased flag, so it will have a
   * valid value before function returns ERROR_VALUE */
  SET_UNERASED;

  if (dbh_locate (dbh, fp) == NULL)
    return ERROR_VALUE;
  dbh->reservedB = CURRENTSEEK;
  if (!CURRENTSEEK)
    return ERROR_VALUE;
  /* don't toggle erased yet! Read the dbh anyways! */
  /*if (ERASED){TOGGLE_ERASE; return ERROR_VALUE;} */

  dbh->bytes_userdata = dbh->newbytes_userdata;
  dbh->branches = dbh->newbranches;
  tmp1 = dbh->key;
  tmp2 = dbh->newkey;
  for (j = 0; j < dbh->head_info->n_limit; j++)
    tmp1[j] = tmp2[j];
  tmp1 = (unsigned char *) dbh->data;
  tmp2 = (unsigned char *) dbh->newdata;
  for (i = 0; i < dbh->newbytes_userdata; i++)
    tmp1[i] = tmp2[i];

  /* must look for CURRENTSEEK, (at dbh->current_seek) if this condition occurs: */
  if (ERASED)
    {
      return ERROR_VALUE;
    }
  return CURRENTSEEK;
}  
/**
 * dbh_regen_sweep:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @Returns: A pointer to the new #DBHashTable.
 *
 * Regenerate the #DBHashTable, eliminating erased records and 
 * optimizing disk access and speed for sweep access.
 * This is done by creating a new #DBHashTable where the physical 
 * structure matches the logical sweep structure. The
 * temporary directory where the new #DBHashTable is created may be set 
 * with dbh_settempdir(). Current #DBHashTable is closed before removed.
 * New #DBHashTable is opened after renamed. 
 *
 **/
DBHashTable *
dbh_regen_sweep (DBHashTable * dbh){
    return dbh_regen(dbh,1);
}

/**
 * dbh_regen_fanout:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @Returns: A pointer to the new #DBHashTable.
 *
 * Regenerate the #DBHashTable, eliminating erased records and 
 * optimizing disk access and speed for fanout access.
 * This is done by creating a new  #DBHashTable where the physical 
 * structure matches the logical fanout structure. The
 * temporary directory where the new #DBHashTable is created may be set 
 * with dbh_settempdir(). Current #DBHashTable is closed before removed.
 * New #DBHashTable is opened after renamed. 
 *
 **/
DBHashTable *
dbh_regen_fanout (DBHashTable * dbh){
    return dbh_regen(dbh,0);
}

/**
 * dbh_find:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @n: Number of branches to ignore on top record.
 * @Returns: 0 on error, byte offset of loaded record otherwise.
 *
 * Find the top level subtree FILE_POINTER for the currently loaded record,
 * but ignoring the last @n branches. 
 * 
 **/
FILE_POINTER
dbh_find (DBHashTable * dbh, int n)
{
  FILE_POINTER fp[3];
  if (dbh == NULL)
    return ERROR_VALUE;
  if (dbh_locateFind (dbh, fp, n) == NULL)
    return ERROR_VALUE;
  if (!CURRENTSEEK)
    return ERROR_VALUE;
  dbh->bytes_userdata = dbh->newbytes_userdata;
  dbh->branches = dbh->newbranches;
  return CURRENTSEEK;
} 
/**
 * dbh_load_address:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @currentseek: A byte offset.
 * @Returns: 0 on error, number of branches otherwise.
 *
 * Load a record from hash table directly from byte offset @currentseek
 *
 **/
unsigned char
dbh_load_address (DBHashTable * dbh, FILE_POINTER currentseek)
{
  TRACE ("dbh_load_address\n");
  unsigned char i;
  if (dbh == NULL)
    return ERROR_VALUE;
  if (currentseek == 0)
    return ERROR_VALUE;
  dbh->reservedB = currentseek;
  for (i = 1; i <= dbh->head_info->n_limit; i++)
    dbh->branch[i - 1] = 0;
  dbh_lock_read (dbh);
  if (!place_fp_at (dbh, currentseek))
    {
      dbh_unlock (dbh);
      return ERROR_VALUE;
    }
  if (!dbh_read (OLD, dbh, 1))
    {
      dbh_unlock (dbh);
      return ERROR_VALUE;
    }
  dbh_unlock (dbh);
  return dbh->branches;
}

/**
 * dbh_load_parent:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @Returns: 0 on error, byte offset of loaded record otherwise.
 *
 * Load the parent of the currently loaded record.
 *
 **/
FILE_POINTER
dbh_load_parent (DBHashTable * dbh)
{
  FILE_POINTER fp[3];
  if (dbh == NULL)
    return ERROR_VALUE;
  if (dbh_locate (dbh, fp) == NULL)
    return ERROR_VALUE;
  if (!CURRENTSEEK || !LASTSEEK)
    return ERROR_VALUE;
  return dbh_load_address (dbh, LASTSEEK);
}

/**
 * dbh_load_child:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @key_index: branch number on which to return the child.
 * @Returns: 0 on error, byte offset of loaded record otherwise.
 *
 * Load the first child of the currently loaded record, on branch identified
 * by @key_index. Since the number of childs (or branches) of each record is 
 * variable, this may be tricky. Top level records have #DBH_KEYLENGTH branches.
 * Lower level records have less. Each byte of a key represents a branch on
 * top level records. 
 *
 **/
FILE_POINTER
dbh_load_child (DBHashTable * dbh, unsigned char key_index)
{
  FILE_POINTER fp[3], child_address;
  if (dbh == NULL)
    return ERROR_VALUE;
  if (dbh_locate (dbh, fp) == NULL)
    return ERROR_VALUE;
  if (!CURRENTSEEK)
    return ERROR_VALUE;
  if (key_index >= dbh->newbranches)
    return ERROR_VALUE;
  child_address = *(dbh->newbranch + key_index);
  return dbh_load_address (dbh, child_address);
} 

/**
 * dbh_erase:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @Returns: 0 on error, 1 otherwise.
 *
 * Mark the record currently loaded into memory as erased. If no record is
 * currently loaded, behaviour is undefined.
 *
 **/
int
dbh_erase (DBHashTable * dbh)
{
  FILE_POINTER currentseek;
  if (dbh == NULL)
    return ERROR_VALUE;
  currentseek = dbh_load (dbh);
  if (!currentseek)		/* will return false if record is already ERASED */
    return ERROR_VALUE;
  TOGGLE_ERASE;			/* set the erased bit on */
  dbh_lock_write (dbh);
  /* set file pointer at record flag byte */
  if (!place_fp_at (dbh, currentseek + 1LL))
    {
      dbh_unlock (dbh);
      return ERROR_VALUE;
    }

  /* write the flag to the file */
  if (write (dbh->fd, &(dbh->flag), 1) != 1)
    {
      dbh_unlock (dbh);
      return ERROR_VALUE;
    }
  /* update file header information */
  dbh->head_info->data_space -= dbh->bytes_userdata;
  dbh->head_info->erased_space += (dbh->bytes_userdata);
  dbh_unlock (dbh);
  dbh_writeheader (dbh);
  return 1;
}

/**
 * dbh_unerase:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @Returns: 0 on error, 1 otherwise.
 *
 * This is the opposite of dbh_erase(). Mark the record currently loaded 
 * into memory as unerased. If no record is currently loaded,
 * behaviour is undefined.
 *
 **/
int
dbh_unerase (DBHashTable * dbh)
{
  FILE_POINTER currentseek, fp[3];
  if (dbh == NULL)
    return ERROR_VALUE;
  /* this will return TRUE if record is already ERASED: */
  if (dbh_locate (dbh, fp) == NULL)
    return ERROR_VALUE;

  currentseek = fp[0];
  if (!currentseek)
    return ERROR_VALUE;
  /* got currentseek, now load it */
  dbh_load_address (dbh, currentseek);	/* found currentseek, now load it */
  if (!ERASED)
    return ERROR_VALUE;		/* hey man, nothing to do */
  TOGGLE_ERASE;
  /* set file pointer at record flag byte */
  dbh_lock_write (dbh);
  if (!place_fp_at (dbh, currentseek + 1LL))
    {
      dbh_unlock (dbh);
      return ERROR_VALUE;
    }
  /* write the flag to the file */
  if (write (dbh->fd, &(dbh->flag), 1) != 1)
    {
      dbh_unlock (dbh);
      return ERROR_VALUE;
    }
  /* update file header information */
  dbh->head_info->data_space += dbh->bytes_userdata;
  dbh->head_info->erased_space -= (dbh->bytes_userdata);
  dbh_unlock (dbh);
  dbh_writeheader (dbh);
  return 1;
} 

/**
 * dbh_prune:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @key: key of top level record of subtree to erase.
 * @subtree_length: number of branches to erase.
 * @Returns: 0 on error, 1 otherwise.
 *
 * Erases a whole subtree from the record currently loaded
 * into memory. Records are not really removed fisically, but 
 * rather marked erased so they may be recovered (if not
 * overwritten later on). Records are permanently removed after
 * #DBHashTable is reconstructed with dbh_regen_sweep() or dbh_regen_fanout().
 *
 **/
int
dbh_prune (DBHashTable * dbh, unsigned char *key,
	   unsigned char subtree_length)
{
  int result;
  dbh->head_info->sweep_erased = 1;
  result = dbh_sweep (dbh, mark_erased, key, NULL, subtree_length);
  dbh->head_info->sweep_erased = 0;
  /* update file header information */
  dbh_writeheader (dbh);
  return result;
}

/**
 * dbh_unprune:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @key: key of top level record of subtree to erase.
 * @subtree_length: number of branches to erase.
 * @Returns: 0 on error, 1 otherwise.
 *
 * Does the opposite of dbh_prune(), marking entire subtree as unerased.
 * May fail to work if records have been overwritten since the 
 * dbh_prune() instruction was issued.
 *
 **/int
dbh_unprune (DBHashTable * dbh, unsigned char *key,
	     unsigned char subtree_length)
{
  int result;
  dbh->head_info->sweep_erased = 1;
  result = dbh_sweep (dbh, mark_unerased, key, NULL, subtree_length);
  dbh->head_info->sweep_erased = 0;
  /* update file header information */
  dbh_writeheader (dbh);
  return result;
}

/**
 * dbh_foreach_sweep:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @operate: A  DBHashFunc() to execute on all records
 * @Returns: 0 on error, 1 otherwise.
 *
 * Apply a function to each member of the hash, following a sweep trajectory.
 * Sweep is done by traversing 
 * the #DBHashTable in a vertical direction through all branches.
 *
 * <para>
 * In order for dbh_foreach_sweep() to be extremely fast, you should 
 * prepare the #DBHashTable for the trajectory with 
 * dbh_regen_sweep() first. This allows for extremely efficient use 
 * of hardware and operating system caches.
 * </para>
 *
 **/
int
dbh_foreach_sweep (DBHashTable * dbh, DBHashFunc operate)
{
  dbh->head_info->dbh_exit = 0;
  if (operate)
    dbh->operate = operate;
  return dbh_newbarre (dbh, NULL, NULL, 0);
}

/**
 * dbh_foreach_fanout:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @operate: A  DBHashFunc() to execute on all records
 * @Returns: 0 on error, 1 otherwise.
 *
 * Apply a function to each member of the hash, following a fanout 
 * trajectory (horizontally through records). dbh_foreach_fanout() is done by 
 * traversing the #DBHashTable in a horizontal direction through all records.
 * <para>
 * In order for dbh_foreach_fanout() to be extremely fast, you should 
 * prepare the #DBHashTable for the trajectory with 
 * dbh_regen_fanout() first. This allows for extremely efficient use 
 * of hardware and operating system caches.
 * </para>
 *
 **/
int
dbh_foreach_fanout (DBHashTable * dbh, DBHashFunc operate)
{
  dbh->head_info->dbh_exit = 0;
  if (operate)
    dbh->operate = operate;
  return dbh_newreversebarre (dbh, NULL, NULL, 0);
}

/**
 * dbh_sweep:
 * @dbh: 	A #DBHashTable pointer (#DBHashTable *).
 * @operate: 	The function to apply to each selected member of the #DBHashTable
 * @key1: 	The key from which to start the sweep or NULL if you don't care. 
 * 	  	Make sure it is a top level node of a subtree with 
 * 	  	dbh_find() first.
 * @key2: 	The key which will trigger an exit condition from the sweep, 
 *        	or NULL if don't care.
 * @ignore_portion: The ignored trailing bytes of key1 which will define the 
 * 	       	magnitud of the subtree to be sweeped, or zero if don't care.
 * @Returns: 	0 on error, 1 otherwise.
 *
 * Apply a function to subtree members of the hash, following a sweep 
 * trajectory (vertically through branches). 
  * <para>
 * In order for dbh_sweep() to be extremely fast, you should 
 * prepare the #DBHashTable for the trajectory with 
 * dbh_regen_sweep() first. This allows for extremely efficient use 
 * of hardware and operating system caches.
 * </para>
*
 **/
int
dbh_sweep (DBHashTable * dbh, DBHashFunc operate,
	   unsigned char *key1,
	   unsigned char *key2, unsigned char ignore_portion)
{
  if (!dbh)
    return ERROR_VALUE;
  if (operate)
    dbh->operate = operate;
  return dbh_newbarre (dbh, key1, key2, ignore_portion);
}

/**
 * dbh_fanout:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @operate: 	The function to apply to each selected member of the #DBHashTable
 * @key1: 	The key from which to start the fanout or NULL if you don't care. 
 * 	  	Make sure it is a top level node of a subtree with 
 * 	  	dbh_find() first.
 * @key2: 	The key which will trigger an exit condition from the sweep, 
 *        	or NULL if don't care.
 * @ignore_portion: The ignored trailing bytes of key1 which will define the 
 * 	       	magnitud of the subtree to be sweeped, or zero if don't care.
 * @Returns: 	0 on error, 1 otherwise.
 *
 * Apply a function to subtree members of the hash, following a fanout
 * trajectory (horizontally through records).
 * <para>
 * In order for dbh_fanout() to be extremely fast, you should 
 * prepare the #DBHashTable for the trajectory with 
 * dbh_regen_fanout() first. This allows for extremely efficient use 
 * of hardware and operating system caches.
 * </para>
 *
 **/int
dbh_fanout (DBHashTable * dbh, DBHashFunc operate,
	    unsigned char *key1,
	    unsigned char *key2, unsigned char ignore_portion)
{
  if (!dbh)
    return ERROR_VALUE;
  if (operate)
    dbh->operate = operate;
  return dbh_newreversebarre (dbh, key1, key2, ignore_portion);
}

/**
 * dbh_exit_sweep:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 *
 * Calling this function from within a #DBHashFunc will cause an 
 * exit of a currently running sweep.
 *
 **/
void
dbh_exit_sweep (DBHashTable * dbh)
{
  if (!dbh)
    {
      fprintf (stderr, "dbh_exitsweep(): %s\n", strerror (EBADF));

      return;
    }

  dbh->head_info->dbh_exit = 1;
}

/**
 * dbh_exit_fanout:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 *
 * Calling this function from within a #DBHashFunc will cause an 
 * exit of a currently running fanout.
 *
 **/
void
dbh_exit_fanout (DBHashTable * dbh)
{
  dbh_exit_sweep(dbh);
}


/**
 * dbh_writeheader:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @Returns: 0 if error, 1 otherwise
 *
 * Write out the DBHashTable header information. It is advisable 
 * to call this function inmediately after creation of a new DBHashTable
 * to force a buffer flush.
 *
 **/
int
dbh_writeheader (DBHashTable * dbh)
{
  if (!dbh)
    {
      fprintf (stderr, "dbh_writeheader(): %s\n", strerror (EBADF));
      return ERROR_VALUE;
    }
  if (strlen (COPYRIGHT) > 127)
    {
      fprintf (stderr, "This should never happen, strlen(\"%s\")>127\n",
	       COPYRIGHT);
      exit (1);
    }
  strcpy ((char *) (dbh->head_info->copyright), COPYRIGHT);

#ifdef TURN
  dbh->head_info->bof = dbh_turnaround (dbh->head_info->bof);
  dbh->head_info->record_length =
    dbh_turnaround (dbh->head_info->record_length);
  dbh->head_info->total_space =
    dbh_turnaround (dbh->head_info->total_space);
  dbh->head_info->data_space = dbh_turnaround (dbh->head_info->data_space);
  dbh->head_info->erased_space =
    dbh_turnaround (dbh->head_info->erased_space);
  dbh->head_info->records = dbh_turnaround (dbh->head_info->records);
  dbh->head_info->fractalidad =
    dbh_turnaround (dbh->head_info->fractalidad);
#endif
  dbh_lock_write (dbh);
  if (!place_fp_at (dbh, 0L))
    return ERROR_VALUE;
  if (write (dbh->fd, dbh->head_info, sizeof (dbh_header_t)) != sizeof (dbh_header_t))
    {
      dbh_unlock (dbh);
      return ERROR_VALUE;
    }
  dbh_unlock (dbh);
#ifdef TURN
  dbh->head_info->bof = dbh_turnaround (dbh->head_info->bof);
  dbh->head_info->record_length =
    dbh_turnaround (dbh->head_info->record_length);
  dbh->head_info->total_space =
    dbh_turnaround (dbh->head_info->total_space);
  dbh->head_info->data_space = dbh_turnaround (dbh->head_info->data_space);
  dbh->head_info->erased_space =
    dbh_turnaround (dbh->head_info->erased_space);
  dbh->head_info->records = dbh_turnaround (dbh->head_info->records);
  dbh->head_info->fractalidad =
    dbh_turnaround (dbh->head_info->fractalidad);
#endif
  return 1;
}

/* This function generates a key that belongs to a finite subset of the cuantified
 * numbers, but which preserves the order of the natural numbers (up to the supreme, 
 * of course) */

/**
 * dbh_orderkey:
 * @key: The address where to put the generated key 
 * @length: The key length
 * @n: The natural number for which to generate the key
 * @base: The number system base to use. This will equal the maximum 
 * 	number of nodes per branch. This ---along with the keylength--- 
 * 	will also define a maximum number of records for the DBHashTable
 * 
 *
 * Obtain a key from a secuential series of natural numbers 
 * (positive integers without zero) which conserves the order of 
 * the natural numbers. This function generates a key that belongs to
 * a finite subset of the cuantified numbers, but which preserves the 
 * order of the natural numbers (up to the supreme, of course)
 *
 **/
void
dbh_orderkey (unsigned char *numero,
	      unsigned char orden, unsigned int n, unsigned char base)
{
  int divisor, i, t;
  double d, b, o;
  if (!n)
    {
      fprintf (stderr,"dbh_genkey: value must be > \n");
      return;
    }

  for (i = 0; i < orden; i++)
    {
      b = base;
      o = orden - 1 - i;
      d = pow (b, o);
      divisor = d;
/*printf("pow(%d,%d) orden=%d, divisor=%d\n",base,orden-1-i,orden,divisor);*/
      numero[i] = (unsigned char) (n / divisor);
      n = n % divisor;
    }
  for (i = 0; i < orden; i++)
    {
      t = numero[i];
      t += 65;
      numero[i] = (unsigned char) t;
    }
  for (i = 0; i < orden; i++)
    {
      if (numero[i] > 'Z')
	{
	  t = numero[i];
	  t += 6;
	  numero[i] = (unsigned char) t;
	}
    }
}

/**
 * dbh_genkey:
 * @key: The address where to put the generated key
 * @length: The key length
 * @n: The natural number from which to generate the key
 *
 * Obtain a key from a secuential series of natural numbers 
 * (positive integers without zero) which does not conserve the order 
 * of the natural numbers, but which are optimized for construction 
 * of a balanced hash tree. These keys are expressed in cuantified 
 * numbers based on the @0 digit.
 *
 **/
void
dbh_genkey (unsigned char *numero, unsigned char orden, unsigned int n)
{
  unsigned char i;
  int t;
  if (!n)
    {
      fprintf (stderr,"dbh_genkey: value must be > \n");
      return;
    }
  dbh_cuenta ((unsigned char *) numero, orden, n);
  for (i = 0; i < orden; i++)
    {
      t = numero[i];
      t += 48;
      numero[i] = (unsigned char) t;
    }
}

/**
 * dbh_genkey2:
 * @key: The address where to put the generated key
 * @length: The key length
 * @n: The natural number from which to generate the key
 *
 * Obtain a key from a secuential series of natural numbers (positive integers 
 * without zero) which does not conserve the order of the natural numbers, 
 * but which are optimized for construction of a balanced hash tree. These
 * keys are expressed in cuantified numbers based on the @A symbol.
 *
 **/
void
dbh_genkey2 (unsigned char *numero, unsigned char orden, unsigned int n)
{
  unsigned char i;
  int t;
  if (!n)
    {
      fprintf (stderr,"dbh_genkey: value must be > \n");
      return;
    }
  dbh_cuenta ((unsigned char *) numero, orden, n);
  for (i = 0; i < orden; i++)
    {
      t = numero[i];
      t += 65;
      numero[i] = (unsigned char) t;
    }
/* for (i=0;i<orden;i++) numero[i] += 65;*/

  for (i = 0; i < orden; i++)
    {
      if (numero[i] > 'Z')
	{
	  t = numero[i];
	  t += 6;
	  numero[i] = (unsigned char) t;
	}
    }
/* for (i=0;i<orden;i++) if (numero[i]>'Z') numero[i] += 6;*/
} 

/**
 * dbh_settempdir:
 * @dbh: A #DBHashTable pointer (#DBHashTable *).
 * @temp_dir: path to temporary directory to use.
 * @Returns: 0 if error, 1 otherwise
 *
 * Sets the temporary directory to be used by dbh_regen_sweep() or 
 * dbh_regen_fanout(). 
 * It is usually best to set temporary directory on the same
 * filesystem device. The default value for the temporary directory is
 * the directory where @dbh is located. To reset to default value, send NULL
 * as the @temp_dir
 *
 **/
int
dbh_settempdir (DBHashTable * dbh, char *dir)
{
  if (!dir){
    dbh->head_info->user_tmpdir=0;
    if (dbh->tmpdir) free (dbh->tmpdir);
    dbh->tmpdir=NULL;
    return 0;
  }
  if (dbh->tmpdir)
    free (dbh->tmpdir);
  dbh->tmpdir = (char *) malloc (strlen (dir) + 1);
  strcpy (dbh->tmpdir, dir);
  dbh->head_info->user_tmpdir=1;
  return 1;
}


/*******************************/
/*     Undocumented Functions: */
/*******************************/

int
dbh_info (DBHashTable * dbh)
{
  if (!dbh)
    {
      fprintf (stderr, "dbh_info(): %s\n", strerror (EBADF));

      return ERROR_VALUE;
    }
  FILE_POINTER eof = place_eof (dbh);
  if (eof < 0)
    return ERROR_VALUE;
  fprintf (stdout, "\nEnd of DBHashTable = %lld\n", (long long) eof);
  fprintf (stdout, "\nDBHashTable dbh_header_t size = %ld",
	   (long) sizeof (dbh_header_t));
  fprintf (stdout, "\ndbh_header:\n\
 version=%s\n\
 keylength=%d\n\
 first record position=%ld\n\
 maximum record length=%ld\n\
 records=%ld\n\
 total_space=%ld\n\
 data_space=%ld\n\
 erased_space=%ld\n\
 format_space=%ld\n\
", dbh->head_info->version, DBH_KEYLENGTH (dbh), (long) dbh->head_info->bof, (long) DBH_MAXIMUM_RECORD_SIZE (dbh), (long) DBH_RECORDS (dbh), (long) DBH_TOTAL_SPACE (dbh), (long) DBH_DATA_SPACE (dbh), (long) DBH_ERASED_SPACE (dbh), (long) DBH_FORMAT_SPACE (dbh));
  return 1;
}   

void
dbh_genkey0 (unsigned char *numero, unsigned char orden, unsigned int n)
{
  if (!n)
    {
      fprintf (stderr,"dbh_genkey: value must be > \n");
      return;
    }
  dbh_cuenta ((unsigned char *) numero, orden, n);
} 

/****************************************************************/
